/*
    saa7110 - Philips SAA7110(A) video decoder driver

    Copyright (C) 1998-2000 Pauline Middelink <middelink@polyware.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/videodev.h>
#include <linux/video_decoder.h>

#ifndef I2C_DRIVERID_SAA7110
#define I2C_DRIVERID_SAA7110	I2C_DRIVERID_EXP0
#endif

#define SAA7110_MAX_INPUT	9	/* 6x CVBS, 3x SVHS on 6 inputs */
#define SAA7110_MAX_OUTPUT	0	/* its a decoder only */

/* Addresses to scan */
static unsigned short normal_i2c[] = {	0x4E, 0x4F, I2C_CLIENT_END };
static unsigned short normal_i2c_range[] = {I2C_CLIENT_END};
I2C_CLIENT_INSMOD;

static struct i2c_client client_template;

static int debug = 0; /* insmod parameter */
MODULE_PARM(debug,"i");

#define dprintk  if (debug) printk

struct saa7110 {
	int		norm;
	int		input;
	int		enable;
	int		bright;
	int		contrast;
	int		hue;
	int		sat;

	unsigned char	regs[54];
};

/* ----------------------------------------------------------------------- */
/* SAA7110 functions							   */
/* ----------------------------------------------------------------------- */
static inline 
void saa7110_write(struct i2c_client *client, u8 reg, u8 data)
{
	struct saa7110* decoder = client->data;

	i2c_smbus_write_byte_data(client,reg,data);
	decoder->regs[reg] = data;
}

static inline 
u8 saa7110_read(struct i2c_client *client)
{
	return i2c_smbus_read_byte(client);
}

static	const unsigned char modes[9][8] = {
/* mode 0 */	{ 0x03, 0xD9, 0x17, 0x40, 0x03, 0x44, 0x75, 0x16 },
/* mode 1 */	{ 0x03, 0xD8, 0x17, 0x40, 0x03, 0x44, 0x75, 0x16 },
/* mode 2 */	{ 0x03, 0xBA, 0x07, 0x91, 0x03, 0x60, 0xB5, 0x05 },
/* mode 3 */	{ 0x03, 0xB8, 0x07, 0x91, 0x03, 0x60, 0xB5, 0x05 },
/* mode 4 */	{ 0x03, 0x7C, 0x07, 0xD2, 0x83, 0x60, 0xB5, 0x03 },
/* mode 5 */	{ 0x03, 0x78, 0x07, 0xD2, 0x83, 0x60, 0xB5, 0x03 },
/* mode 6 */	{ 0x83, 0x59, 0x17, 0x42, 0xA3, 0x44, 0x75, 0x12 },
/* mode 7 */	{ 0x83, 0x9A, 0x17, 0xB1, 0x13, 0x60, 0xB5, 0x14 },
/* mode 8 */	{ 0x83, 0x3C, 0x27, 0xC1, 0x23, 0x44, 0x75, 0x21 } };

static
int saa7110_selmux(struct i2c_client *client)
{
	struct saa7110* decoder = client->data;
	const unsigned char* ptr = modes[decoder->input];

	/*
	 * Note, switching channels looses the special case
	 * for BW signals (see determine_norm(). Most drivers
	 * however will do an autodetect /AFTER/ switching
	 * channels so it won't matter much.
	 */
	saa7110_write(client,0x06,ptr[0]);	/* Luminance control	*/
	saa7110_write(client,0x20,ptr[1]);	/* Analog Control #1	*/
	saa7110_write(client,0x21,ptr[2]);	/* Analog Control #2	*/
	saa7110_write(client,0x22,ptr[3]);	/* Mixer Control #1	*/
	saa7110_write(client,0x2C,ptr[4]);	/* Mixer Control #2	*/
	saa7110_write(client,0x30,ptr[5]);	/* ADCs gain control	*/
	saa7110_write(client,0x31,ptr[6]);	/* Mixer Control #3	*/
	saa7110_write(client,0x21,ptr[7]);	/* Analog Control #2	*/

	return 0;
}

static
int determine_norm(struct i2c_client* client)
{
	struct saa7110* decoder = client->data;
	int	status;

	/*
	 * For the detection process we need to be in NTSC mode.
	 */
	if (decoder->norm != VIDEO_MODE_NTSC) {
		saa7110_write(client,0x06,0x03);	/* bypass active */
		saa7110_write(client,0x0D,0x86);
		saa7110_write(client,0x0F,0x50);
		saa7110_write(client,0x11,0x2C);
		saa7110_write(client,0x2E,0x81);
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(HZ);
	}
	
	/* mode changed, start automatic detection */
	status = saa7110_read(client);
dprintk(KERN_DEBUG "%s: status1=%02X\n",client->name,status);
	if ((status & 3) == 0) {
		/*
		 * in BW mode we can bypass the chromatrap and have
		 * a slightly sharper image.
		 */
		saa7110_write(client,0x06,0x83);
		if (status & 0x20) {
			dprintk(KERN_DEBUG "%s: norm=bw60\n",client->name);
			saa7110_write(client,0x0F,0x50);
			saa7110_write(client,0x2E,0x81);
			return /*VIDEO_MODE_NTSC*/4;
		}
		dprintk(KERN_DEBUG "%s: norm=bw50\n",client->name);
		saa7110_write(client,0x0F,0x10);
		saa7110_write(client,0x2E,0x9A);
		return /*VIDEO_MODE_PAL*/3;
	}

	/*
	 * In COLOR mode we can bypass the chromatrap only for
	 * the Video-S modes, since they give us a seperate chroma.
	 * This results in a slightly sharper image.
	 */
	saa7110_write(client,0x06,modes[decoder->input][0]);
	saa7110_write(client,0x0D,0x86);
	if (status & 0x20) {	/* 60Hz */
		dprintk(KERN_DEBUG "%s: norm=ntsc\n",client->name);
		saa7110_write(client,0x0F,0x50);
		saa7110_write(client,0x11,0x2C);
		saa7110_write(client,0x2E,0x81);
		return VIDEO_MODE_NTSC;
	}

	/* 50Hz -> PAL/SECAM */
	saa7110_write(client,0x0F,0x10);
	saa7110_write(client,0x11,0x59);
	saa7110_write(client,0x2E,0x9A);

	current->state = TASK_UNINTERRUPTIBLE;
	schedule_timeout(HZ/4);

	status = saa7110_read(client);
dprintk(KERN_DEBUG "%s: status2=%02X\n",client->name,status);
	if ((status & 0x03) == 0x01) {
		dprintk(KERN_DEBUG "%s: norm=secam\n",client->name);
		saa7110_write(client,0x0D,0x87);
		return VIDEO_MODE_SECAM;
	}
	dprintk(KERN_DEBUG "%s: norm=pal\n",client->name);
	return VIDEO_MODE_PAL;
}

static
int saa7110_attach(struct i2c_adapter *adap, int addr,
			  unsigned short flags, int kind)
{
	static	unsigned char initseq[] = {
	/* decoder */
	     0,	0x4C,	/* 00 - IDEL */
		0x3C,	/* 01 - HSYB 50hz */
		0x0D,	/* 02 - HSYS 50hz */
		0xEF,	/* 03 - HCLB 50hz */
		0xBD,	/* 04 - HCLS 50hz */
		0xF0,	/* 05 - HPHI 50hz */
		0x03,	/* 06 - BYPS=0, PREF=0, BPSS=0, CORI=0, APER=3 */
		0x00,	/* 07 - HUEC */
		0xF8,	/* 08 - CKTQ */
		0xF8,	/* 09 - CKTS */
		0x60,	/* 0A - PLSE */
		0x60,	/* 0B - SESE */
		0x00,	/* 0C - COLO=0, LFIS=0 */
		0x84,	/* 0D - VTRC=1, RTSE=0, HRMV=1, SSTB=0, SECS=0 */
		0x18,	/* 0E - HPLL=0, OEHV=1, OEYC=1, CHRS=0, GPSW=0 */
		0x10,	/* 0F - AUFD=0, FSEL=0, SXCR=0, SCEN=0, YDEL=0 */
		0x00,	/* 10 - HRFS=0, VNOI=0 */
		0x59,	/* 11 - CHCV */
		0x40,	/* 12 - SATN */
		0x46,	/* 13 - CONT */
		0x42,	/* 14 - HS6B 60hz */
		0x1A,	/* 15 - HS6S 60hz */
		0xFF,	/* 16 - HC6B 60hz */
		0xDA,	/* 17 - HC6S 60hz */
		0xF0,	/* 18 - HP6I */
		0x8B,	/* 19 - BRIG */
	/* reserved */
		0x00,	/* 1A - reserved */
		0x00,	/* 1B - reserved */
		0x00,	/* 1C - reserved */
		0x00,	/* 1D - reserved */
		0x00,	/* 1E - reserved */
		0x00,	/* 1F - reserved */
	/* front-end */
		0xD9,	/* 20 - AIND=6, FUSE=1, AINS=1 */
		0x17,	/* 21 - VBCO=0, MS34=0, MX24=1, MS24=0, REFS=7 */
		0x40,	/* 22 - GACO=1, CSEL=0, YSEL=0, MUYC=0, CLTS=0, MX34=0 */
		0x41,	/* 23 - CLL21 */
		0x80,	/* 24 - CLL22 */
		0x41,	/* 25 - CLL31 */
		0x80,	/* 26 - CLL32 */
		0x4F,	/* 27 - HOLD=0, GASL=1, GAI2=0F */
		0xFE,	/* 28 - WIPE */
		0x01,	/* 29 - SBOT */
		0xCF,	/* 2A - IWIP=3, GAI3=0F */
		0x0F,	/* 2B - IGAI=0, GAI4=0F */
		0x03,	/* 2C - CLS4=0, CLS3=0, CLS2=0, TWO3=1, TWO2=1 */
		0x01,	/* 2D - IVAL */
		0x9A,	/* 2E - VBPS */
		0x03,	/* 2F - VBPR */
		0x40,	/* 30 - WISL=1, GAS3=0, GAD3=0, GAS2=0, GAD2=0 */
		0x75,	/* 31 - AOSL=3, WIRS=1, WRSE=1, SQPB=0, VBLK=0, PULI=1 */
		0x01,	/* 32 - WVAL */
		0x8C,	/* 33 - OFTS=1, CHSB=0, CAD3=0, CAD2=0 */
		0x03};	/* 34 - MUD2=0, GUDL=3 */
	struct saa7110		*decoder;
	struct i2c_client	*client;
	int			rv;

	client = kmalloc(sizeof(*client)+sizeof(*decoder),GFP_KERNEL);
	if (!client)
		return -ENOMEM;		
        memcpy(client,&client_template,sizeof(*client));
        client->adapter = adap;
        client->addr = addr;
	client->data = decoder = (struct saa7110*)(client+1);

	/* clear our private data */
	memset(decoder, 0, sizeof(struct saa7110));
	decoder->norm = VIDEO_MODE_PAL;
	decoder->input = 0;
	decoder->enable = 1;
	decoder->bright = 32768;
	decoder->contrast = 32768;
	decoder->hue = 32768;
	decoder->sat = 32768;

	/* the SAA does autoincrement on continued writes... */
	rv = i2c_master_send(client, initseq, sizeof(initseq));
	if (rv != sizeof(initseq)) {
		printk(KERN_ERR "%s_attach: init status %d\n", client->name, rv);
		/* stop further output */
		saa7110_write(client,0x0E,0x00);
		kfree(client);
		return -1;
	}
	memcpy(decoder->regs,initseq+1,sizeof(decoder->regs));

	/* REFS off, Clamp on */
	saa7110_write(client,0x21,0x16);

	/* note StatusByteSelect=0 in the init-seq */
	dprintk(KERN_DEBUG "%s_attach: chip version %x\n", client->name, saa7110_read(client));
	saa7110_write(client,0x0D,0x86);

	/* setup and implicit mode 0 select has been performed */
	printk(KERN_DEBUG "saa7110: initialized\n");

	i2c_attach_client(client);
	return 0;
}

int saa7110_probe(struct i2c_adapter *adap)
{	
	if (adap->id == (I2C_ALGO_BIT | I2C_HW_B_BT848))
		return i2c_probe(adap, &addr_data, saa7110_attach);
	return 0;
}

static
int saa7110_detach(struct i2c_client *client)
{
	int err;

	dprintk(KERN_DEBUG "%s_detach\n",client->name);

	/* stop further output */
	saa7110_write(client,0x0E,0x00);

	/* Try to detach the client from i2c space */
	if ((err = i2c_detach_client(client))) {
		dprintk(KERN_DEBUG "%s_detach: Client deregistration failed, client not detached.\n",client->name);
		return err;
	}

	kfree(client);
	return 0;
}

static
int saa7110_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	struct saa7110* decoder = client->data;
	int	v;

	switch (cmd) {
	 case DECODER_GET_CAPABILITIES:
		{
			struct video_decoder_capability *dc = arg;
			dc->flags = VIDEO_DECODER_PAL
				  | VIDEO_DECODER_NTSC
				  | VIDEO_DECODER_SECAM
				  | VIDEO_DECODER_AUTO
				  | VIDEO_DECODER_CCIR
#ifdef VIDEO_DECODER_RAW
				  | VIDEO_DECODER_RAW
#endif
				  ;
			dc->inputs = SAA7110_MAX_INPUT;
			dc->outputs = SAA7110_MAX_OUTPUT;
		}
		break;

	 case DECODER_GET_STATUS:
		{
			int status;
			int res = 0;

			status = saa7110_read(client);
			if (status & 0x40)
				res |= DECODER_STATUS_GOOD;
			if (status & 0x03)
				res |= DECODER_STATUS_COLOR;

			if (status & 0x20) {
				/* 60Hz; always NTSC */
				res |= DECODER_STATUS_NTSC;
			}
			/* 50Hz; could be PAL or SECAM */
			else if (decoder->regs[0x0D] & 0x01) {
				if (status & 0x02)
					res |= DECODER_STATUS_SECAM;
				else
					res |= DECODER_STATUS_PAL;
			}
			else {
				if (status & 0x02)
					res |= DECODER_STATUS_PAL;
				else
					res |= DECODER_STATUS_SECAM;
			}
			*(int*)arg = res;
		}
		break;

	 case DECODER_SET_NORM:
		v = *(int*)arg;
		if (decoder->norm != v) {
			decoder->norm = v;
			/* force bypass in a known mode */
			saa7110_write(client,0x06,modes[decoder->input][0]);
			switch (v) {
			 case VIDEO_MODE_NTSC:
				dprintk(KERN_DEBUG "%s: set norm=ntsc\n",client->name);
				saa7110_write(client,0x0D,0x86);
				saa7110_write(client,0x0F,0x50);
				saa7110_write(client,0x11,0x2C);
				saa7110_write(client,0x2E,0x81);
				break;
			 case VIDEO_MODE_PAL:
				dprintk(KERN_DEBUG "%s: set norm=pal\n",client->name);
				saa7110_write(client,0x0D,0x86);
				saa7110_write(client,0x0F,0x10);
				saa7110_write(client,0x11,0x59);
				saa7110_write(client,0x2E,0x9A);
				break;
			 case VIDEO_MODE_SECAM:
				dprintk(KERN_DEBUG "%s: set norm=secam\n",client->name);
				saa7110_write(client,0x0D,0x87);
				saa7110_write(client,0x0F,0x10);
				saa7110_write(client,0x11,0x59);
				saa7110_write(client,0x2E,0x9A);
				break;
			 case VIDEO_MODE_AUTO:
				decoder->norm = determine_norm(client);
				dprintk(KERN_DEBUG "%s: set determine_norm() gueses %d\n",client->name,decoder->norm);
				*(int*)arg = decoder->norm;
				break;
			 default:
				return -EPERM;
			}
		}
		break;

#ifdef VIDEO_DECODER_RAW
	 case DECODER_SET_MODE:
		v = *(int*)arg;
		switch (v) {
		 case 0: /* YUV */
			saa7110_write(client,0x31, v ? decoder->reg[0x31] & 0xF7);
			break;
		 case 1: /* RAW */
			/* its not very good RAW, but the output of AD2/AD3 */
			saa7110_write(client,0x31, v ? decoder->reg[0x31] | 0x08);
			break;
		 case 2: /* TTX */
			/* fall through */
		 default:
			return -EINVAL;
		}
		break;
#endif

	 case DECODER_SET_INPUT:
		v = *(int*)arg;
		if (v<0 || v>=SAA7110_MAX_INPUT)
			return -EINVAL;
		if (decoder->input != v) {
			decoder->input = v;
			saa7110_selmux(client);
		}
		break;

	 case DECODER_SET_OUTPUT:
		v = *(int*)arg;
		/* not much choice of outputs */
		if (v<0 || v>=SAA7110_MAX_OUTPUT)
			return -EINVAL;
		break;

	 case DECODER_ENABLE_OUTPUT:
		v = *(int*)arg;
		if (decoder->enable != v) {
			decoder->enable = v;
			saa7110_write(client,0x0E, v ? 0x18 : 0x00);
		}
		break;

	 case DECODER_SET_PICTURE:
		{
			struct video_picture *pic = arg;

			if (decoder->bright != pic->brightness) {
				/* We want 0 to 255 we get 0-65535 */
				decoder->bright = pic->brightness;
				saa7110_write(client,0x19,decoder->bright >> 8);
			}
			if (decoder->contrast != pic->contrast) {
				/* We want 0 to 127 we get 0-65535 */
				decoder->contrast = pic->contrast;
				saa7110_write(client,0x13,decoder->contrast >> 9);
			}
			if (decoder->sat != pic->colour) {
				/* We want 0 to 127 we get 0-65535 */
				decoder->sat = pic->colour;
				saa7110_write(client,0x12,decoder->sat >> 9);
			}
			if (decoder->hue != pic->hue) {
				/* We want -128 to 127 we get 0-65535 */
				decoder->hue = pic->hue;
				saa7110_write(client,0x07,(decoder->hue>>8)-128);
			}
		}
		break;

	 case DECODER_DUMP:
		for (v=0; v<sizeof(decoder->regs); v+=16) {
			int j;
			dprintk(KERN_DEBUG "%s: %03x ",client->name,v);
			for (j=v; j<v+16 && j<sizeof(decoder->regs); j++) {
				dprintk(" %02x",decoder->regs[j]);
			}
			dprintk("\n");
		}
		break;

	 default:
		dprintk(KERN_DEBUG "unknown saa7110_command??(%x)\n",cmd);
		return -EINVAL;
	}
	return 0;
}

static
void saa7110_inc_use(struct i2c_client *client)
{
	MOD_INC_USE_COUNT;
}

static
void saa7110_dec_use(struct i2c_client *client)
{
	MOD_DEC_USE_COUNT;
}
 
static struct i2c_driver driver = {
	name:           "i2c saa7110 driver",
	id:             I2C_DRIVERID_SAA7110,
	flags:          I2C_DF_NOTIFY,
	attach_adapter: saa7110_probe,
	detach_client:  saa7110_detach,
	command:        saa7110_command,
	inc_use:        saa7110_inc_use,
	dec_use:        saa7110_dec_use
};

static struct i2c_client client_template =
{
        name:   "saa7110",
        driver: &driver
};

int __init saa7110_init(void)
{
	i2c_add_driver(&driver);
	return 0;
}

void __exit saa7110_exit(void)
{
	i2c_del_driver(&driver);
}

module_init(saa7110_init);
module_exit(saa7110_exit);
